/*
 * ffwd.c
 *
 * This is for the pure-C feed-forward program "ff" by Jeff Bilmes.
 * It adds the following features:
 *
 *  - Read a neural net definition from an MLP3-style wts file 
 *    as written by Quicknet.
 *  - Write output in lna format (or rapbin, if compiled -DWRITE_RAPBIN)
 *  - Read input patterns in online feature format
 *  - Handle assembling input patterns into context_window blocks
 *
 * 1999oct21 dpwe@icsi.berkeley.edu
 * $Header: /u/drspeech/src/ffwd/RCS/ffwd.c,v 1.4 2000/08/18 06:23:34 dpwe Exp $
 */

#include <stdio.h>
#include <stdlib.h> /* for atoi */
#include <math.h>
#include <assert.h>
#include "ff.h"
#include "io.h"

#ifdef WRITE_RAPBIN

#define out_open       rapbin_open_w
#define out_close      rapbin_close
#define out_write      rapbin_write
#define out_write_eos  rapbin_write_eos
#define OPFMT          "rapbin"

#else /* !WRITE_RAPBIN - must be LNA */

#define out_open       lna_open_w
#define out_close      lna_close
#define out_write      lna_write
#define out_write_eos  lna_write_eos
#define OPFMT          "lna"

#endif /* RAPBIN/LNA */

int norm_ftrs(const float *src, float *dst, int width, int nframes, 
	      int sskip, int dskip, 
	      const float *bias, const float *scale) 
{   /* normalize according to bias and scale read from file */
    /* skip is pts to skip at end of each frame (bias) */
    int i,j;
    const float *bptr, *sptr;
    for (i = 0; i < nframes; ++i) {
	bptr = bias;
	sptr = scale;
	for (j = 0; j < width; ++j) {
	    *dst++ = (*sptr++) * (*src++ + *bptr++);
	}
	dst += dskip;
	src += sskip;
    }
}

int onl_norm_ftrs(const float *src, float *dst, int width, int nframes, 
		  int sskip, int dskip,
		  float *bias, float *scale, 
		  float alpha_m, float alpha_v) 
{   /* online normalization of features */
    int i, j;
    float *bptr, *sptr;
    for (i = 0; i < nframes; ++i) {
	bptr = bias;
	sptr = scale;
	for (j = 0; j < width; ++j) {
	    double mean = - *bptr;
	    double var = 1 / (*sptr * *sptr);
	    double x = *src++;

	    mean = (1 - alpha_m) * mean + alpha_m * x;
	    x -= mean;
	    var = (1 - alpha_v) * var + alpha_v * x * x;

	    *bptr++ = - mean;
	    *sptr = 1 / sqrt(var);		

	    *dst++ = x * *sptr++;
	}
	src += sskip;
	dst += dskip;
    }
}

int ninp = 0;
int nhid = 0;
int nout = 0;
int width = 0;
int cw = 9;
int bsize = 16;
char *ifile = NULL;
char *ofile = "-";	/* stdout by default */
float bval = DEFBIASUNITVAL;
char *wfile = NULL;
char *nfile = NULL;

int onlnorm = 0;
float alpha_m = 0.005;
float alpha_v = 0.005;

void usage(char *name)
{
  fprintf(stderr,"usage: %s [OPTIONS]\n",name);
  fprintf(stderr,"where [OPTIONS] include:\n");  
  fprintf(stderr,"  -ninp i                 : number of input units.\n");
  fprintf(stderr,"  -nhid h                 : number of hidden units.\n");
  fprintf(stderr,"  -nout o                 : number of output units.\n");
  fprintf(stderr,"  -width w                : input ftr width.\n");
  fprintf(stderr,"  -cw p                   : input window context width.\n");
  fprintf(stderr,"  -bsize b                : bunch size\n");
  fprintf(stderr,"  -onlnorm                : apply online feature normalization\n");
  fprintf(stderr,"  -alpha_m a              : update constant for ftr means (0.005)\n");
  fprintf(stderr,"  -alpha_v b              : update constant for ftr vars (0.005)\n");
  fprintf(stderr,"  -ifile file             : input ftr file (online_ftr format)\n");
  fprintf(stderr,"  -ofile file             : output file (" OPFMT " format)\n");
  fprintf(stderr,"  -wfile file             : file containing net definition (QN MLP3 wts fmt)\n");
  fprintf(stderr,"  -nfile file             : file defining norms for ftrs (qnnorm fmt)\n");
}

void
parse_args(int argc,char*argv[])
{
  int i;
  if (argc == 1) {
    usage(argv[0]);
    exit(-1);
  }
  for (i=1;i<argc;i++) {
    if (argv[i][0] == '-') {
      if (!strcmp(&(argv[i][1]),"ninp")) {
	ninp = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"nhid")) {
	nhid = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"nout")) {
	nout = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"width")) {
	width = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"cw")) {
	cw = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"bsize")) {
	bsize = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"bval")) {
	bval = atof(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"ifile")) {
	ifile = argv[++i];
      } else if (!strcmp(&(argv[i][1]),"ofile")) {
	ofile = argv[++i];
      } else if (!strcmp(&(argv[i][1]),"wfile")) {
	wfile = argv[++i];
      } else if (!strcmp(&(argv[i][1]),"nfile")) {
	nfile = argv[++i];
      } else if (!strcmp(&(argv[i][1]),"onlnorm")) {
	onlnorm = 1;
      } else if (!strcmp(&(argv[i][1]),"alpha_m")) {
	alpha_m = atof(argv[++i]);;
      } else if (!strcmp(&(argv[i][1]),"alpha_v")) {
	alpha_v = atof(argv[++i]);;
      } else  {
        fprintf(stderr,"Error: Unknown or invalid argument (%s)\n",argv[i]);
	exit(-1);
      }
    }
  }
  if (ninp <= 0 || nhid <= 0 || nout <= 0 || width <= 0 || cw <= 0 || \
      bsize <= 0 || ifile == NULL || ofile == NULL || wfile == NULL) {
    /* Note: it's OK for nfile to be null. */
    fprintf(stderr,"Error with arguments\n");
    fprintf(stderr, "ninp=%d nhid=%d nout=%d width=%d cw=%d bsize=%d\n", 
	    ninp, nhid, nout, width, cw, bsize);
    fprintf(stderr, "ifile=%s\n", ifile?ifile:"(null)");
    fprintf(stderr, "ofile=%s\n", ofile?ofile:"(null)");
    fprintf(stderr, "wfile=%s\n", wfile?wfile:"(null)");
    fprintf(stderr, "nfile=%s\n", nfile?nfile:"(null)");
    usage(argv[0]);
    exit(-1);
  }
  if (ninp != cw * width) {
      fprintf(stderr, "ninp (%d) != cw (%d) * width (%d)!!\n", 
	      ninp, cw, width);
      exit(-1);
  }
}

int
main(int argc,char *argv[]) 
{
  /* weights */
  float *who;  /* hidden to output weights */
  float *wih;  /* hidden to output weights */
  
  /* patterns */
  float *ipats; /* input patterns */

  /* scratch & misc. */
  float *hpats; /* hidden layer */
  float *opats; /* output layer */

  int j, nread, nwrit;

  FILE *ipfile, *opfile;
  
  float *opp, *ipp, *ipp2;
  float *bias = NULL, *scale = NULL;

  int done = 0, eos = 0;

  parse_args(argc,argv);

  who = lmalloc(nout*(nhid+1)*sizeof(float));
  wih = lmalloc(nhid*(ninp+1)*sizeof(float));
  ipats = lmalloc(bsize*(ninp+1)*sizeof(float));
  hpats = lmalloc(bsize*(nhid+1)*sizeof(float));
  opats = lmalloc(bsize*nout*sizeof(float));

  if (readmlp3(wfile, ninp, nhid, nout, wih, who) == 0) {
      fprintf(stderr, "Error reading wts file '%s'\n", wfile);
      exit(1);
  }

  if ( (ipfile = onlftrs_open_r(ifile)) == NULL) {
      fprintf(stderr, "Unable to open input file '%s'\n", ifile);
      exit(1);
  }

  if ( (opfile = out_open(ofile, nout)) == NULL) {
      fprintf(stderr, "Unable to open output file '%s'\n", ofile);
      exit(1);
  }

  if ( nfile ) {
      bias  = lmalloc(width*sizeof(float));
      scale = lmalloc(width*sizeof(float));
      if (readnorms(nfile, width, bias, scale) == 0) {
	  fprintf(stderr, "Unable to read norms file '%s'\n", nfile);
	  exit(1);
      }
  }

  if (onlnorm) {
      if (!bias) {
	  fprintf(stderr, "Must specify a norms file to inialize -onlnorm\n");
	  exit(1);
      }
  }

  bias_matrix(hpats,bsize,nhid+1,bval);
  bias_matrix(ipats,bsize,ninp+1,bval);

  /* outer segment loop */
  while (!done) {

      /* prepad the input */
      /* we read into the final part of the input pattern buffer, 
	 which is where the block-iteration loop expects to find them. */
      ipp = ipats + bsize*(ninp+1) - (cw-1)*width - 1;
      nread = 0;
      while (!done && nread < cw - 1) {
	  eos = onlftrs_read(ipfile, ipp, width);
	  if (onlnorm) onl_norm_ftrs(ipp, ipp, width, 1, 1, 1, 
				     bias, scale, alpha_m, alpha_v);
	  else if (bias) norm_ftrs(ipp, ipp, width, 1, 1, 1,
				   bias, scale);
	  if (eos < 0) {
	      /* Hit EOF probably (could be a read error) */
	      if (nread > 0) {
		  fprintf(stderr, "WARN: unexpected EOF in first CW\n");
	      }
	      done = 1;
	  } else {
	      if (eos) {
		  fprintf(stderr, "Unexpected EOS before first CW filled\n");
		  done = 1;
	      }
	      ipp += width;
	      ++nread;
	  }
      }

      nwrit = 0;

      while (!done && !eos) {

	  /* First row of block copies repeated patterns from end of buffer */
	  ipp2 = ipats + bsize*(ninp+1) - (cw-1)*width - 1;
	  /* but we're copying *to* the beginning */
	  ipp = ipats;

	  /* build a whole buffer full */
	  nread = 0;
	  while (!eos && nread < bsize) {
	      /* first copy cw-1 patterns */
	      memcpy(ipp, ipp2, (cw-1)*width*sizeof(float));
	      /* then read in one new pattern */
	      eos = onlftrs_read(ipfile, ipp+(cw-1)*width, width);
	      if (eos < 0) {
		  fprintf(stderr, "WARN: EOF without EOS mark\n");
		  done = 1;
	      }
	      if (onlnorm) onl_norm_ftrs(ipp+(cw-1)*width, ipp+(cw-1)*width, 
					 width, 1, 1, 1, 
					 bias, scale, alpha_m, alpha_v);
	      else if (bias) norm_ftrs(ipp+(cw-1)*width, ipp+(cw-1)*width, 
				       width, 1, 1, 1, bias, scale);
	      if (!done) {
		  ++nread;
		  /* point to last few patterns ready for next copy */
		  ipp2 = ipp+width;
		  /* .. and point to start of next row (skip bias col) */
		  ipp += ninp+1;
	      }
	  }

	  if (nread > 0) {

	      /* Now do the fwd pass */
	      bunched_mlp2_forwprop_sm(ninp,nhid,nout,nread,
				       who,wih,
				       ipats,
				       hpats,opats);

	      /* write the outputs */
	      opp = opats;
	      for (j = 0; j < nread; ++j) {
		  out_write(opfile, opp, nout);
		  opp += nout;
		  ++nwrit;
	      }
	  }
      }

      if (nwrit) {
	  out_write_eos(opfile);
      }

  }	/* end of outer segment loop */

  out_close(opfile);

  exit(0);
}
