/*
 * io.c
 * 
 * Input/output routines for the 'ffwd' pure-C forward pass code.
 *
 * 1999nov08 dpwe@icsi.berkeley.edu
 * $Header: /u/drspeech/src/ffwd/RCS/io.c,v 1.1 1999/11/09 05:07:06 dpwe Exp $
 */

#include <stdio.h>
#include <stdlib.h> /* for atoi */
#include <math.h>
#include <assert.h>
#include <malloc.h>
#include "ff.h"	    /* for lmalloc */
#include "io.h"

int readmlp3(char *fname, int ninp, int nhid, int nout, 
	     float *ihwts, float *howts)
{   /* <fname> is the name of an mlp3 ascii wts file of 
       <ninp>:<nhid>:<nout>.  <ihwts> points to a pre-allocated 
       buffer of <nhid> rows of <ninp>+1 cols (bias as last column);
       <howts> points to <nout> rows fo <nhid+1> cols. 
       Return 1 on success, 0 on fail. */
    FILE *f = fopen(fname, "r");
    char s[64];
    int nw, i, j, nread;
    float *fp, x;

    if (f == NULL) {
	fprintf(stderr, "Couldn't open wts file '%s'\n", fname);
	return 0;
    }

    /* First comes "weigvec" */
    if (fscanf(f, "%s %d", s, &nw) != 2) {
	goto rderr;
    }
    if (strcmp(s, "weigvec") != 0) {
	fprintf(stderr, "expecting 'weigvec' got '%s'\n", s);
	goto rderr;
    }
    if (nw != ninp * nhid) {
	fprintf(stderr, "nwts = %d != ni x nh = %d x %d = %d\n", 
		nw, ninp, nhid, ninp*nhid);
	goto rderr;
    }

    /* Read the IH weights */
    fp = ihwts;
    nread = 0;
    for (i = 0; i < nhid; ++i) {
	for (j = 0; j < ninp; ++j) {
	    if (fscanf(f, "%f", &x) != 1) {
		fprintf(stderr, "Error reading IH wt (after %d)\n", nread);
		goto rderr;
	    }
	    *fp++ = x;
	    ++nread;
	}
	/* skip the bias wt at the end of each line */
	++fp;
    }

    /* Read the HO wts */
    if (fscanf(f, "%s %d", s, &nw) != 2) {
	goto rderr;
    }
    if (strcmp(s, "weigvec") != 0) {
	fprintf(stderr, "expecting 'weigvec' got '%s'\n", s);
	goto rderr;
    }
    if (nw != nhid * nout) {
	fprintf(stderr, "nwts = %d != nh x no = %d x %d = %d\n", 
		nw, nhid, nout, nhid*nout);
	goto rderr;
    }
    fp = howts;
    nread = 0;
    for (i = 0; i < nout; ++i) {
	for (j = 0; j < nhid; ++j) {
	    if (fscanf(f, "%f", &x) != 1) {
		fprintf(stderr, "Error reading HO wt (after %d)\n", nread);
		goto rderr;
	    }
	    *fp++ = x;
	    ++nread;
	}
	/* skip the bias wt at the end of each line */
	++fp;
    }

    /* Now read the biasvecs - hidden first */
    if (fscanf(f, "%s %d", s, &nw) != 2) {
	goto rderr;
    }
    if (strcmp(s, "biasvec") != 0) {
	fprintf(stderr, "expecting 'biasvec' got '%s'\n", s);
	goto rderr;
    }
    if (nw != nhid) {
	fprintf(stderr, "nbias = %d != nh = %d\n", 
		nw, nhid);
	goto rderr;
    }
    fp = ihwts + ninp;	/* start at end of first row */
    nread = 0;
    for (j = 0; j < nhid; ++j) {
	if (fscanf(f, "%f", &x) != 1) {
	    fprintf(stderr, "Error reading H bias (after %d)\n", nread);
	    goto rderr;
	}
	/* Negate bias wts to match to -1 root */
	*fp++ = -x;
	/* skip to next row */
	fp += ninp;
	++nread;
    }

    /* Output layer biasvec */
    if (fscanf(f, "%s %d", s, &nw) != 2) {
	goto rderr;
    }
    if (strcmp(s, "biasvec") != 0) {
	fprintf(stderr, "expecting 'biasvec' got '%s'\n", s);
	goto rderr;
    }
    if (nw != nout) {
	fprintf(stderr, "nbias = %d != nh = %d\n", 
		nw, nout);
	goto rderr;
    }
    fp = howts + nhid;	/* start at end of first row */
    nread = 0;
    for (j = 0; j < nout; ++j) {
	if (fscanf(f, "%f", &x) != 1) {
	    fprintf(stderr, "Error reading O bias (after %d)\n", nread);
	    goto rderr;
	}
	/* Negate bias wts to match to -1 root */
	*fp++ = -x;
	/* skip to next row */
	fp += nhid;
	++nread;
    }

    /* All read OK */
    fclose(f);
    return 1;

 rderr:
    fprintf(stderr, "Error reading wts file '%s'\n", fname);
    fclose(f);
    return 0;
}

int readnorms(char *fname, int npts, float *bias, float *scale) {
    /* Read norms file, as written by qnnorm (or us) */
    FILE *f = fopen(fname, "r");
    char s[64];
    int nw, i, j, nread;
    float *fp, x;

    if (f == NULL) {
	fprintf(stderr, "Couldn't open norms file '%s'\n", fname);
	return 0;
    }

    /* First comes "vec nn" */
    if (fscanf(f, "%s %d", s, &nw) != 2) {
	goto rderr;
    }
    if (strcmp(s, "vec") != 0) {
	fprintf(stderr, "expecting 'vec' got '%s'\n", s);
	goto rderr;
    }
    if (nw != npts) {
	fprintf(stderr, "vec's n = %d != %d\n", 
		nw, npts);
	goto rderr;
    }
    /* Read the biases */
    fp = bias;
    nread = 0;
    for (i = 0; i < npts; ++i) {
	if (fscanf(f, "%f", &x) != 1) {
	    fprintf(stderr, "Error reading means (after %d)\n", nread);
	    goto rderr;
	}
	/* negate so we add it later */
	*fp++ = -x;
	++nread;
    }

    /* All over again for the scales */
    if (fscanf(f, "%s %d", s, &nw) != 2) {
	goto rderr;
    }
    if (strcmp(s, "vec") != 0) {
	fprintf(stderr, "expecting 2nd 'vec' got '%s'\n", s);
	goto rderr;
    }
    if (nw != npts) {
	fprintf(stderr, "2nd vec's n = %d != %d\n", 
		nw, npts);
	goto rderr;
    }
    fp = scale;
    nread = 0;
    for (i = 0; i < npts; ++i) {
	if (fscanf(f, "%f", &x) != 1) {
	    fprintf(stderr, "Error reading sds (after %d)\n", nread);
	    goto rderr;
	}
	/* recip so we mult by it later */
	/* *fp++ = 1.0/x; */
	/* no, data is stored as recip */
	*fp++ = x;
	++nread;
    }
    
    fclose(f);
    return npts;

 rderr:
    fprintf(stderr, "Error reading norms file '%s'\n", fname);
    fclose(f);
    return 0;
}

int writenorms(char *fname, int npts, float *means, float *sds) {
    /* Write out a norms file from the supplied data */
    FILE *fp;
    int i;
    if (strcmp(fname, "-") == 0) {
	/* use stdout */
	fp = stdout;
    } else {
	fp = fopen(fname, "w");
    }
    if (fp == NULL) {
	fprintf(stderr, "Couldn't write norms file '%s'\n", fname);
	return 0;
    }
    fprintf(fp, "vec %d\n", npts);
    for(i=0; i<npts; ++i) {
	fprintf(fp, "%f\n", means[i]);
    }
    fprintf(fp, "vec %d\n", npts);
    for(i=0; i<npts; ++i) {
	/* data is stored as invsds */
	fprintf(fp, "%f\n", 1.0/sds[i]);
    }
    if (fp != stdout) fclose(fp);
    return npts;
}

static void swab_fbuf(float *buf, int n) {
    /* Swap bytes in a buffer of floats, if needed to make them bigendian */
#ifndef WORDS_BIGENDIAN
  unsigned int *ptr = (unsigned int *)buf;
  unsigned int d;
  int i;

  assert(sizeof(unsigned int) == sizeof(float));
  for (i=0; i<n; ++i) {
    d = *ptr;
    *ptr++ = ((d & 0x000000FF) << 24) | ((d & 0x0000FF00) << 8) | ((d & 0x00FF0000) >> 8) | (0x000000FF & ((d & 0xFF000000) >> 24));
  }
#endif /* WORDS_BIGENDIAN */
}

static void swab_ibuf(int *buf, int n) {
    /* Swap bytes in a buffer of ints, if needed to make them bigendian */
#ifndef WORDS_BIGENDIAN
  unsigned int *ptr = (unsigned int *)buf;
  unsigned int d;
  int i;

  assert(sizeof(unsigned int) == sizeof(int));
  for (i=0; i<n; ++i) {
    d = *ptr;
    *ptr++ = ((d & 0x000000FF) << 24) | ((d & 0x0000FF00) << 8) | ((d & 0x00FF0000) >> 8) | (0x000000FF & ((d & 0xFF000000) >> 24));
  }
#endif /* WORDS_BIGENDIAN */
}

FILE *onlftrs_open_r(char *name) {
    /* Open the input online features stream */
    FILE *f;
    if (strcmp(name, "-") == 0) {
	/* use stdin */
	f = stdin;
    } else {
	f = fopen(name, "rb");
    }
    if (f == NULL) {
	fprintf(stderr, "onlftrs_open_r: error opening '%s'\n", name);
	return NULL;
    }

    return f;
}

void onlftrs_close(FILE *f) {
    if (f == stdin) {
	fflush(f);
    } else {
	fclose(f);
    }
}

int onlftrs_read(FILE *fp, float *buf, int nftrs)
{   /* Read a frame of online features from the input.  
       Normally, return 0.  On error, return -1.  At EOS, return 1. */
    /* OLFF is a flag word/byte, then bigendian floats */
    int flagbytes = 1;	/* or 4 for cambridge version */
    int flag = 0;
    int nr;

    /* read flag */
    if ( fread(&flag, flagbytes, 1, fp) == 0) {
	return(-1);
    }
    if (flagbytes == 4) {
	swab_ibuf(&flag, 1);
    }

    /* read ftrs */
    if ( (nr = fread(buf, sizeof(float), nftrs, fp)) < nftrs) {
	fprintf(stderr, "Error reading onl ftr frame - got %d of %d\n", 
		nr, nftrs);
	return(-1);
    }
    swab_fbuf(buf, nftrs);
    
    /* return code based on flag */
    return (flag == 0)?0:1;
}

static int rapbin_frnum = 0;
    
FILE *rapbin_open_w(char *name, int width)
{   /* Open the named file for rapbin output */
    FILE *f;
    if (strcmp(name, "-") == 0) {
	/* use stdout */
	f = stdout;
    } else {
	f = fopen(name, "wb");
    }
    if (f == NULL) {
	fprintf(stderr, "rapbin_open_w: error opening '%s'\n", name);
	return NULL;
    }

    /* Otherwise, write header i.e. width of stream */
    swab_ibuf(&width, 1);
    fwrite(&width, sizeof(int), 1, f);
    swab_ibuf(&width, 1);

    rapbin_frnum = 0;

    return f;
}

void rapbin_close(FILE *f) {
    if (f == stdout) {
	fflush(f);
    } else {
	fclose(f);
    }
}

int rapbin_write(FILE *f, float *vals, int nftrs)
{   /* Write a frame of output data in rapbin format. */

    /* Write frame number */
    swab_ibuf(&rapbin_frnum, 1);
    fwrite(&rapbin_frnum, sizeof(int), 1, f);
    swab_ibuf(&rapbin_frnum, 1);
    ++rapbin_frnum;

    /* Write data */
    swab_fbuf(vals, nftrs);
    fwrite(vals, sizeof(int), nftrs, f);
    swab_fbuf(vals, nftrs);

    return 1;
}

int rapbin_write_eos(FILE *f) {    
    /* Write EOS mark */
    int eosmark = -1;

    swab_ibuf(&eosmark, 1);
    fwrite(&eosmark, sizeof(int), 1, f);
    fflush(f);
    swab_ibuf(&eosmark, 1);
    rapbin_frnum = 0;

    return 1;
}

/*
 * LNA file writing
 */

unsigned char *lna_buf = NULL;
int lna_width = 0;
int lna_buf_full = 0;

FILE *lna_open_w(char *name, int width)
{   /* Open the named file for lna output */
    FILE *f;
    if (strcmp(name, "-") == 0) {
	/* use stdout */
	f = stdout;
    } else {
	f = fopen(name, "wb");
    }
    if (f == NULL) {
	fprintf(stderr, "lna_open_w: error opening '%s'\n", name);
	return NULL;
    }

    /* allocate buffer for caching output frames */
    if (width > lna_width) {
	if (lna_buf)	free(lna_buf);
	lna_buf = (unsigned char*)malloc(width + 1);   /* flag byte too */
    }
    lna_width = width;
    /* clear flag byte */
    lna_buf[0] = 0;
    /* clear semaphore that there is data to be written */
    lna_buf_full = 0;
    
    return f;
}

void lna_close(FILE *f) {
    assert(lna_buf_full == 0);
    if (f == stdout) {
	fflush(f);
    } else {
	fclose(f);
    }
}

#define LNA8_VERY_SMALL (1e-37)
#define LNA8_SCALE (24.0)
 
int lna_write(FILE *f, float *vals, int nftrs)
{   /* Write a frame of output data in lna format. */
    int i,x;
    assert(nftrs == lna_width);
    /* We can't write this frame until we know if it's the last 
       one or not.  But we can write the previous frame, if any */
    if (lna_buf_full) {
	fwrite(lna_buf, 1, nftrs+1, f);
    }
    /* Store these values to be written next time */
    for (i = 0; i < nftrs; ++i) {
	x = (int) floor( - LNA8_SCALE * log( vals[i] + LNA8_VERY_SMALL ) ) ;
	if (x>255) x = 255;
	if (x<0)   x = 0;
	lna_buf[i+1] = x;
    }
    lna_buf_full = 1;

    return 1;
}

int lna_write_eos(FILE *f) {    
    /* Write EOS mark i.e. set the flag & write the last frame*/
    assert(lna_buf_full);
    lna_buf[0] = 128;
    fwrite(lna_buf, 1, lna_width+1, f);
    fflush(f);
    lna_buf[0] = 0;
    lna_buf_full = 0;
    return 1;
}

